# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Created on Mar 12, 2014

author: jakeret
"""
from __future__ import print_function, division, absolute_import, unicode_literals


class UnsupportedCompilerException(Exception):
    pass

class UnsupportedFeatureException(Exception):
    pass